@@ -361,3 +361,49 @@ function AFPercent_Keystroke( nDec, sepStyle )
         event.rc = !isNaN(event.change) || event.change == "." || event.change == ","
     }
 }
+
+app.popUpMenuEx = function() {
+    return app.okular_popUpMenuEx(arguments);
+}
+
+app.popUpMenu = function() {
+    // Convert arguments like this:
+    //   app.popUpMenu(["Fruits","Apples","Oranges"], "-","Beans","Corn");
+    // into this:
+    //   app.popUpMenuEx(
+    //     {cName:"Fruits", oSubMenu:[
+    //       {cName:"Apples"},
+    //       {cName:"Oranges"}
+    //     ]},
+    //     {cName:"-"},
+    //     {cName:"Beans"},
+    //     {cName:"Corn"}
+    //   );
+    function convertArgument(arg) {
+        var exArguments = [];
+
+        for (element of arg) {
+            var newElement = null;
+
+            if (Array.isArray(element) && element.length > 0) {
+                newElement = {
+                    cName: element[0],
+                    oSubMenu: convertArgument(element.slice(1))
+                };
+            } else if (!Array.isArray(element)) {
+                newElement = {
+                    cName: element
+                };
+            }
+
+            if (newElement !== null)
+                exArguments.push(newElement);
+        }
+
+        return exArguments;
+    }
+
+    var exArguments = convertArgument(arguments);
+    var result =  app.okular_popUpMenuEx(exArguments);
+    return result;
+}
\ No newline at end of file
