@@ -1335,6 +1335,10 @@ Okular::TextPage *PDFGenerator::textPage(Okular::TextRequest *request)
     std::vector<std::unique_ptr<Poppler::TextBox>> textList;
     double pageWidth, pageHeight;
     userMutex()->lock();
+    if (request->shouldAbortExtraction()) {
+        userMutex()->unlock();
+        return nullptr;
+    }
     std::unique_ptr<Poppler::Page> pp = pdfdoc->page(page->number());
     if (pp) {
         TextExtractionPayload payload(request);
@@ -1705,7 +1709,7 @@ bool PDFGenerator::exportTo(const QString &fileName, const Okular::ExportFormat
 inline void append(Okular::TextPage *ktp, const QString &s, double l, double b, double r, double t)
 {
     //    qWarning(PDFDebug).nospace() << "text: " << s << " at (" << l << "," << t << ")x(" << r <<","<<b<<")";
-    ktp->append(s, new Okular::NormalizedRect(l, t, r, b));
+    ktp->append(s, Okular::NormalizedRect(l, t, r, b));
 }
 
 Okular::TextPage *PDFGenerator::abstractTextPage(const std::vector<std::unique_ptr<Poppler::TextBox>> &text, double height, double width, int rot)
