@@ -22,7 +22,8 @@
 
 #include "config-okular.h"
 #if HAVE_X11
-#include <QX11Info>
+#include <KX11Extras>
+#include <private/qtx11extras_p.h>
 #endif
 #if HAVE_DBUS
 #include <QDBusConnectionInterface>
@@ -103,7 +104,11 @@ static bool attachExistingInstance(const QStringList &paths, const QString &seri
     const QString pattern = QStringLiteral("org.kde.okular-");
     const QString myPid = QString::number(qApp->applicationPid());
     QScopedPointer<QDBusInterface> bestService;
-    const int desktop = KWindowSystem::currentDesktop();
+#if HAVE_X11
+    const int desktop = KX11Extras::currentDesktop();
+#else
+    const int desktop = 0;
+#endif
 
     // Select the first instance that isn't us (metric may change in future)
     for (const QString &service : services) {
@@ -111,7 +116,7 @@ static bool attachExistingInstance(const QStringList &paths, const QString &seri
             bestService.reset(new QDBusInterface(service, QStringLiteral("/okularshell"), QStringLiteral("org.kde.okular")));
 
             // Find a window that can handle our documents
-            const QDBusReply<bool> reply = bestService->call(QStringLiteral("canOpenDocs"), paths.count(), desktop);
+            const QDBusReply<bool> reply = bestService->call(QStringLiteral("canOpenDocs"), (int)paths.count(), desktop);
             if (reply.isValid() && reply.value()) {
                 break;
             }
